/* 
 * xsvfparser.c
 * 
 * readin and parse a XSVF file and output interpretation to buffers to be loaded to the FX2
 * 
 * Originally by: Peter Civitarese And James Gilmore
 * Heaviliy modified by: Patrick McNeil and Nathan Friess to work with the FX2
 * Modified to run on OS X by: Andrew Stagg
 *
 * SOURCE(s) 
 * 
 * Format and Descritpion of XSVF Xilinx Commands were referenced from Xilinx XAPP058 (x3.1) June 25, 2004
 * 
 */

/* INCLUDES */
#include <CoreFoundation/CoreFoundation.h>
#include "xsvfparser.h"

/* GLOBALS */
FILE *in;
uchar cmd;
uchar val[MAXLEN + 1];
uchar bigdata[108000]; //big array to hold the data
uchar xrepeat_times;
uchar xruntest_time[4];
int   xsir_length;
uchar len[4];
int   length, total_length;
uchar TDOmask[MAXLEN];
int	  bytecnt;

/*Buffer for commands sent to FX2*/
uchar buff[6];

/*
* read a byte form the file 
*/
int readByte(uchar *ret)
{
	uchar r_byte;
	size_t result;

	result = fread(&r_byte, 1, 1, in);
	if(result != 1) {
		//check EOF
		if(feof(in) == 0) {
			return 1;
		}
	}
	else
		bytecnt++;

	if (ret)
		*ret = r_byte;

	return 0;
}

/*
* return an int value for a 4 char array
*/
int calcLength(uchar * ptr)
{
	int lvalue;
	int i;

	lvalue = 0;
	for(i=0; i<3; i++) {
		lvalue |= ptr[i];
		lvalue <<= 8;
	}
	lvalue |= ptr[i];

	return lvalue;
}

/*
* Read in the TDI value specified by length
*/
void read_tdi() {
	int i;
	for(i=length-1; i>=0; i--) {
		uchar c;
		if (readByte(&c))
			return;
		val[i] = flipbits(c);
	}
}

/*
* Fill a bufer with configuration data from the XSVF to be sent to the FX2
*/
void fillbigdata() {
	int i;
	for(i=total_length+length-1; i>=total_length; i--) {
		uchar c;
		if (readByte(&c))
			return;
		bigdata[i] = flipbits(c);
	}
	total_length+=length;
}

/*
* puts total_length into buff for SIR
*/
void length_to_buff() {
	int fixedlength=0; //, temp=0;
	void *temp2 = &fixedlength;
	
	//remember to convert to 32bit words
	fixedlength=(int)ceil((double)total_length/4);

	//flip the bytes
	buff[4] = *((char *)temp2+0);
	buff[5] = *((char *)temp2+1);
}

/*
* flip the bits of a byte (because the data in XSVF needs to be sent LSB first)
*/
uchar flipbits(uchar c) {
	int i;
	unsigned char d = 0;

	for (i = 0; i < 7; i++) {
		d |= (c & 1);
		c >>= 1;
		d <<= 1;
	}

	d |= (c & 1);

	return d;
}

/*
* parser - takes in a filename of an XSVF and parses and sends the data to the FX2
*/
int parse(char *file)
{

	int result;
	int i;

	// open file
	in = fopen(file, "rb");
	if(in == NULL) {
		return 1;
	}

	// while we have not reached the EOF continue reading
	while(feof(in) == 0)
	{

		// read in a command, note:ERROR CHECKING done in readByte()
		if (readByte(&cmd)) {
			fclose(in);
			return 1;
		}
		
		// evaluate command
		switch(cmd)
		{

		case XCOMPLETE:
			/*
			* XCOMPLETE <no args>
			*
			* End of XSVF file reached
			*/

			fclose(in);
			return 0;

		case XSDRSIZE:
			/* 
			* XSDRSIZE length<4 bytes>
			*
			* specifies length of data to come
			*/

			// read length<4 bytes>, also stick it in the buffer, for the next command
			for(i=0; i<4; i++) {
				if (readByte(len+i)) {
					fclose(in);
					return 1;
				}
			}
			
			// calculate length as an int - our length specifies the number of bytes to read not "bits"
			length = calcLength(len);
			length = (int)ceil((double)length/8);
			break;

		case XSIR:
			/*
			* XSIR length<l byte> TDIValue<"length" bits>
			*
			* Shift instruction into instruction register
			*/

			// read length<1 byte>
			if (readByte(val)) {
				fclose(in);
				return 1;
			}

			// calculate xsir_length (number of bytes to read, length specifies "bits")
			xsir_length = (int)val[0];
			xsir_length = (int)ceil((double)xsir_length/8);

			// read TDIValue<"length" bits>
			for(i=0; i<xsir_length; i++) {
				if (readByte(buff+i)) {
					fclose(in);
					return 1;
				}
			}
			
			// actually send the command here, unless it's CFG_IN, since we need to know the data length first!
			if(buff[0]!=INST_CFG_IN) {
				if(buff[0]==INST_JSTART)
					if(send_command(buff)){fclose(in); return 1;}
			}
			break;

		case XSDR:
			/*
			* XSDR TDIValue<"length" bits>
			*
			* Shift data into data register
			*/
			
			// read TDIValue<"length" bits>
			read_tdi();

			// send SIR command - CFG_IN and then send the data
			total_length=length;
			length_to_buff();
			
			// only send CFG_IN data, here we are actually sending CFG_IN data to FX2
			if(buff[0]==INST_CFG_IN) {
				if(send_command(buff)){fclose(in); return 1;}
				if (send_data(val,total_length)){fclose(in); return 1;}
			}
			break;

		case XSDRTDO:
			/*
			* XSDRTDO TDIValue<"length" bits> TDOExpected<"length" bits>
			*
			* Same as XSDR except that there is an expected TDO, which we will be ignoring
			*/

			// read TDIValue<"length" bits>
			read_tdi();
			
			// read TDO expected but don't do anything with it, we just need to parse it out
			for(i=0; i<length; i++) {
				uchar c;
				readByte(&c);
			}

			// send SIR command - CFG_IN and then send the data
			total_length=length;
			length_to_buff();
			
			// only send CFG_IN data
			if(buff[0]==INST_CFG_IN) {
				if(send_command(buff)){fclose(in); return 1;}
				if (send_data(val,total_length)){fclose(in); return 1;}
			}
			break;

		case XSDRB:
			/*
			* XSDRB TDIValue<"length" bits>
			*
			* Go to the shift-DR state and shift in the TDI value. 
			* Continue to stay in the shift-DR state at the  end of the operation. 
			* No comparison of TDO value with the last specified TDOExpected is  performed.
			*/

			// for this, the data doesn't get actually sent until XSDRE, so we need to add up the length and data pieces...
			total_length=0;
			fillbigdata();

			break;

		case XSDRC:
			/*
			* XSDRC TDIValue<"length" bits>
			*
			* Shift in the TDI value. Continue to stay in the shift-DR state at the end of the operation. 
			* No comparison of TDO value with the last specified TDOExpected is performed. 
			*/

			// read TDIValue<"length" bits> into the big buffer
			fillbigdata();

			break;

		case XSDRE:
			/*
			* XSDRE TDIValue<"length" bits>
			*
			* Shift in the TDI value. At the end of the operation, go to the XENDDR state. 
			* No comparison of  TDO value with the last specified TDOExpected is performed.
			*/

			// read TDIValue<"length" bits>
			fillbigdata();

			length_to_buff();

			// send SIR command - CFG_IN and then send the data
			if(buff[0]==INST_CFG_IN) {
				if(send_command(buff)){fclose(in); return 1;}
				if (send_data(bigdata,total_length)){fclose(in); return 1;}
			}
			break;

		/* 
		* NOTE: These commands can be safely ignored but still need to be parsed out
		*/
		case XTDOMASK:
			for(i=0; i<length; i++) {if (readByte(TDOmask+i)) {fclose(in); return 1;}}
			break;

		case XRUNTEST:
			for(i=0; i<4; i++) {if (readByte(xruntest_time+i)) {fclose(in); return 1;}} 
			break;

		case XREPEAT:
			if (readByte(&xrepeat_times)) {
				fclose(in);
				return 1;
			}
			break;
			
		case XSTATE:
			if (readByte(val)) {
				fclose(in);
				return 1;
			}
			break;
			
		case XENDIR:
			if (readByte(val)) {
				fclose(in);
				return 1;
			}
			break;
			
		case XENDDR:
			if (readByte(val)) {
				fclose(in);
				return 1;
			}
			break;
			
		/* 
		* NOTE: testing has shown the rest of these commands are never present in our XSVF files
		* so we just added these casses for completeness
		*/
		case XSETSDRMASKS:
		case XSDRINC:
		case XSDRTDOB:
		case XSDRTDOC:
		case XSDRTDOE:
			return 1; // error
			
		// anything else? well this should never happen
		default:
			return 1; // error

		} //END switch

	} //END while(!eof(in))

	// close file
	result = fclose(in);
	if(result != 0) {
		return 1;
	}

	//success
	return 0;
}
