/*
 * usb-skel, interrupts.c
 *
 * By Nathan Friess and Patrick McNeil
 * April 12, 2005
 *
 * The entry point for the firmware.
 */

#include "fx2.h"
#include "fx2regs.h"
#include "usbdescr.h"

// Used by the epcs macro 
const char code  EPCS_Offset_Lookup_Table[] =
{
   0,    // EP1OUT
   1,    // EP1IN
   2,    // EP2OUT
   2,    // EP2IN
   3,    // EP4OUT
   3,    // EP4IN
   4,    // EP6OUT
   4,    // EP6IN
   5,    // EP8OUT
   5,    // EP8IN
};

// Macro for generating the address of an endpoint's control and status register (EPnCS)
#define epcs(EP) (EPCS_Offset_Lookup_Table[(EP & 0x7E) | (EP > 128)] + 0xE6A1)

// Flags set by ISRs
volatile unsigned char ep1_out_rdy;
volatile unsigned char ep1_in_rdy;
volatile unsigned char ep2_rdy;

// USB Descriptors
extern xdata unsigned int DeviceDescr;
extern xdata unsigned int DeviceQualifierDescr;
extern xdata unsigned int HighSpeedConfigDescr;
extern xdata unsigned int FullSpeedConfigDescr;
extern xdata unsigned int StringDscr0;
extern xdata unsigned int StringDscr1;
extern xdata unsigned int StringDscr2;

#define MSB(val) ((unsigned char)(((unsigned int)val >> 8) & 0xFF))
#define LSB(val) ((unsigned char)((unsigned int)val & 0xFF))

void ISR_USB_sudav() interrupt
{

	// Request type
	switch(SETUPDAT[1])
	{

	case SC_GET_DESCRIPTOR:

		// Type
		switch(SETUPDAT[3])
		{
		case GD_DEVICE:
			SUDPTRH = MSB(&DeviceDescr);
			SUDPTRL = LSB(&DeviceDescr);
			break;

		case GD_DEVICE_QUALIFIER:
			SUDPTRH = MSB(&DeviceQualifierDescr);
			SUDPTRL = LSB(&DeviceQualifierDescr);
			break;

		case GD_CONFIGURATION:
			SUDPTRH = MSB(&HighSpeedConfigDescr);
			SUDPTRL = LSB(&HighSpeedConfigDescr);
			break;

		case GD_OTHER_SPEED_CONFIGURATION:	// USB 1 Configuration?
			SUDPTRH = MSB(&FullSpeedConfigDescr);
			SUDPTRL = LSB(&FullSpeedConfigDescr);
			break;

		case GD_STRING:
			// Index
			switch (SETUPDAT[2])
			{
			case 0:
				SUDPTRH = MSB(&StringDscr0);
				SUDPTRL = LSB(&StringDscr0);
				break;

			case 1:
				SUDPTRH = MSB(&StringDscr1);
				SUDPTRL = LSB(&StringDscr1);
				break;

			case 2:
				SUDPTRH = MSB(&StringDscr2);
				SUDPTRL = LSB(&StringDscr2);
				break;
			}	
			break;

		default:
			// Set stall... unexpected problem.
			EP0CS |= 0x01;
		}
		break;

	case SC_SET_DESCRIPTOR:
		// Read descriptor data... but do what with it?
		break;

	case SC_GET_INTERFACE:
		// Get alternate setting for an interface.
		EP0BUF[0] = 0;
		EP0BCH = 0;
		EP0BCL = 1;
		break;

	case SC_SET_INTERFACE:
		// Change alternate settings for an interface.  Use as you see fit.
		// E.g. Sampling rate for an audio device.

		// Reset all endpoints...
		// Clear HSNAK
		_asm 
			nop
		_endasm;
		break;

	case SC_SET_CONFIGURATION:
		// Set configuration number.  Use as you see fit.
		// E.g. High power and lower power modes.
		_asm 
			nop
		_endasm;
		break;

	case SC_GET_CONFIGURATION:
		// Return the current config number.
		EP0BUF[0] = 1; // The default
		EP0BCH = 0;
		EP0BCL = 1;
		break;

	case SC_GET_STATUS:	

		// Type
		switch(SETUPDAT[0])
		{
		case GS_DEVICE:
			EP0BUF[0] = 0; // Not self-powered, No remote wakeup
			EP0BUF[1] = 0;
			EP0BCH = 0;
			EP0BCL = 2;
			break;

		case GS_ENDPOINT:
			EP0BUF[0] = epcs(SETUPDAT[4]) & 0x01; // Stalled?
			EP0BUF[1] = 0;
			EP0BCH = 0;
			EP0BCL = 2;
			break;

		case GS_INTERFACE:
			EP0BUF[0] = 0; // Always return zeros
			EP0BUF[1] = 0;
			EP0BCH = 0;
			EP0BCL = 2;
			break;

		}
		break;

	case SC_SET_FEATURE:
		// For setting remote wake-up or stall bits.
		_asm 
			nop
		_endasm;

		break;

	case SC_CLEAR_FEATURE:
		// For clearing remote wake-up or stall bits.
		_asm 
			nop
		_endasm;

		break;
	}

	// Acknowledge handshake phase of device request
	EP0CS |= 0x80;

	EXIF &= 0xef;

	USBIRQ = 1;
}

void ISR_USB_sof() interrupt
{
	EXIF &= 0xef;
	USBIRQ = 0x02;
}

void ISR_USB_sutok() interrupt 
{
	EXIF &= 0xef;
	USBIRQ = 0x04;
}

void ISR_USB_susp() interrupt
{
	EXIF &= 0xef;
	USBIRQ = 0x08;
}

void ISR_USB_ures() interrupt
{
	EXIF &= 0xef;
	USBIRQ = 0x10;
}

void ISR_USB_highspeed() interrupt
{
	EXIF &= 0xef;
	USBIRQ = 0x20;
}        

void ISR_USB_ep0ack() interrupt
{
	EXIF &= 0xef;
	USBIRQ = 0x40;
}

void ISR_USB_stub() interrupt
{
	EXIF &= 0xef;
	USBIRQ = 0x80;
}                     

void ISR_USB_ep0in() interrupt
{
	EXIF &= 0xef;
	EPIRQ = 0x01;
}

void ISR_USB_ep0out() interrupt
{
	EXIF &= 0xef;
	EPIRQ = 0x02;
}

// Endpoint 1 IN is ready for data.
void ISR_USB_ep1in() interrupt
{
	ep1_in_rdy = 1;

	EXIF &= 0xef;
	EPIRQ = 0x04;
}

// Endpoint 1 OUT data has been acknowledged.
void ISR_USB_ep1out() interrupt
{
	ep1_out_rdy = 1;

	EXIF &= 0xef;
	EPIRQ = 0x08;
}
